import numpy as np
from joblib import Parallel, delayed, cpu_count
import numba as nb
import threading

# JIT-accelerated functions
@nb.jit(nb.float64[:, :](nb.float64[:, ::1], nb.float64[:, ::1]), fastmath=True, nopython=True)
def fastDot(X, Y):
    return np.dot(X, Y)

@nb.jit(nb.float64[:,:](nb.float64[:], nb.float64[:]), nopython=True)
def fastBuf(principalComponentColumn, eigenVectorsRow):
    return np.outer(principalComponentColumn, eigenVectorsRow)[::-1]

@nb.jit(nb.float64[:,:](nb.float64[:,:]), nopython=True)
def fastCovariance(X):
    return np.cov(X)

@nb.jit(nopython=True)
def fastMean(X):
    return X.mean()

# Helper functions for multi-threading
def getColumn(s, K, m):
    return s[m:K+m]

def laggingFunction(idx, LaggedMatrix, s, K):
    LaggedMatrix[:, idx] = getColumn(s, K, idx)

def calcRC(antiDiagonal):
    return fastMean(antiDiagonal)

def accelerated_ssa(signal, L=10, numComp=10):
    '''
    Accelerated Singular Spectrum Analysis (SSA) using Numba & Joblib.

    Parameters:
        signal (array-like): The input signal.
        L (int): Window length for embedding.
        numComp (int): Number of components to reconstruct.

    Returns:
        np.ndarray: Reconstructed components as columns.
    '''
    N = len(signal)
    K = N - L + 1

    # Embedding using multi-threading
    X = np.zeros((K, L))
    threads = []
    for m in range(L):
        threads.append(threading.Thread(target=laggingFunction, args=(m, X, signal, K)))
        threads[m].start()
    for m in range(L):
        threads[m].join()

    # Eigenvalue decomposition
    eigenValues, eigenVectors = np.linalg.eigh(fastCovariance(X.T) / K)
    idx = np.argsort(eigenValues)[::-1]
    eigenValues, eigenVectors = eigenValues[idx], eigenVectors[:, idx]

    # Convert to C-contiguous arrays for speed
    eigenVectors, X = np.ascontiguousarray(eigenVectors), np.ascontiguousarray(X)

    # Compute principal components
    PC = fastDot(X, eigenVectors)

    # Reconstructed components
    RC = np.zeros((N, numComp), dtype=np.float64)
    for i in range(numComp):
        buf = fastBuf(PC[:, i], eigenVectors[:, i].T)
        Diag0, Diag1 = -buf.shape[0] + 1, buf.shape[1]
        RC[:, i] = Parallel(n_jobs=cpu_count())(delayed(calcRC)(buf.diagonal(j)) for j in range(Diag0, Diag1))

    return RC



import numpy as np

def classic_ssa(signal, L=10, numComp=10):
    '''
    Classic Singular Spectrum Analysis (SSA) based on:
    https://www.kaggle.com/code/jdarcy/introducing-ssa-for-time-series-decomposition
    
    Parameters:
        signal (array-like): The input signal.
        L (int): Window length for embedding.
        numComp (int): Number of components to reconstruct.

    Returns:
        np.ndarray: Reconstructed components as columns.
    '''
    N = len(signal)  # Lag length
    K = N - L + 1  # Embedded signal length

    # Embedding
    X = np.array([signal[i:L+i] for i in range(0, K)]).T

    # Decomposition by Singular Value Decomposition
    U, Sigma, VT = np.linalg.svd(X)

    # Reconstructed components
    RC = np.zeros((N, numComp))  
    for i in range(numComp):
        buff = Sigma[i] * np.outer(U[:, i], VT[i, :])  # Buffer matrix
        buff = buff[::-1]  # Flip for easier anti-diagonal selection
        RC[:, i] = [buff.diagonal(j).mean() for j in range(-buff.shape[0] + 1, buff.shape[1])]
    
    return RC



import numpy as np
from .classic import classic_ssa
from .accelerated import accelerated_ssa

def SSA(signal, lag, numComp, method="classic"):
    """
    Wrapper function to perform Singular Spectrum Analysis (SSA).
    
    Parameters:
        signal (numpy.ndarray): The input time-series signal.
        lag (int): Window length for embedding.
        numComp (int): Number of components to reconstruct.
        method (str): Either 'classic' or 'accelerated' SSA.

    Returns:
        np.ndarray: Reconstructed components as columns.
    """
    assert isinstance(signal, np.ndarray) and signal.ndim == 1, "signal must be 1D and of type ndarray"
    assert isinstance(lag, int) and 2 <= lag < len(signal) // 2, "Lag must be between 2 and len(signal)//2"
    assert isinstance(numComp, int) and 1 <= numComp <= lag, "Number of components must be between 1 and lag"
    
    if method == "classic":
        return classic_ssa(signal, lag, numComp)
    elif method == "accelerated":
        return accelerated_ssa(signal, lag, numComp)
    else:
        raise ValueError("Options are either 'classic' or 'accelerated'")



from setuptools import setup, find_packages

setup(
    name="ssa_tools",
    version="0.1.1",
    author="Mehieddine Derbas",
    author_email="mehyeddin.derbass@gmail.com",
    description="A Python library for Singular Spectrum Analysis (SSA)",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/TinoDerb/ssa_tools",  # Replace with your GitHub URL
    packages=find_packages(),
    install_requires=[
        "numpy",
        "joblib",
        "numba",
        "tqdm"
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.7",
    license="MIT",  # Explicitly mention the license here
)



import numpy as np
import pytest
from ssa_tools.core import SSA

def test_ssa_output_classic():
    signal = np.sin(np.linspace(0, 10, 100))
    components = SSA(signal, lag=20, numComp=5, method="classic")
    assert components.shape == (100, 5)

def test_invalid_lag_classic():
    signal = np.sin(np.linspace(0, 10, 100))
    with pytest.raises(AssertionError):
        SSA(signal, lag=1, numComp=5, method="classic")

def test_ssa_output_accelerated():
    signal = np.sin(np.linspace(0, 10, 100))
    components = SSA(signal, lag=20, numComp=5, method="accelerated")
    assert components.shape == (100, 5)

def test_invalid_lag_accelerated():
    signal = np.sin(np.linspace(0, 10, 100))
    with pytest.raises(AssertionError):
        SSA(signal, lag=1, numComp=5, method="accelerated")
